;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; First-stage bootloader for FAT12 floppy
; Chris Giese <geezer@execpc.com>, http://my.execpc.com/~geezer
;
; I, the copyright holder of this work, hereby release it into the
; public domain. This applies worldwide. If this is not legally possible:
; I grant any entity the right to use this work for any purpose,
; without any conditions, unless such conditions are required by law.
;
; 10 Dec. 2014:
; - Default name of second stage changed from "LOAD.BIN" to "STARTUP.BIN"
;   for compatability with Alexei Frounze's BootProg
; - Default load address changed from 10000h (64K) to 8000h (32K)
; - It now loads the proper number of sectors, even if sectors_per_cluster
;   is >1 (tested w/ 1.68 meg DMF floppy; 4 sectors/cluster, 3 sectors/FAT,
;   1 sector for root directory)
; - Cluster values are now tested to make sure they're valid (>1 and <0xFF7).
;   If not, a blinking white on blue 'S' (fileSystem error) is displayed.
; - Should work with any sensible bytes-per-sector value (up to 4096);
;   though this has not been well-tested
; - Now using '286 and '386 instructions (no test for 32-bit CPU, though)
;
; 26 Nov. 2002:
; - Initial release
; - Finds file "LOAD.BIN" stored in the root directory, loads it into
;   memory, and jumps to it. "LOAD.BIN" need not be stored in physically
;   contiguous disk sectors (i.e. it need not be "defragged").
;
; xxx - want an assemble-time check to make sure SS_ADDR >= 7E00h
;
; BUILD:
;	nasm -f bin -o fat12.bin fat12.asm
;
; DOS INSTALL; 1.44 meg floppies only:
;	partcopy fat12.bin  0 200 -f0
; -or-
;	debug fat12.bin
;	-w 100 0 0 1	('-' is DEBUG prompt. Type ? for help.)
;	-q
;
; DOS INSTALL; floppies other than 1.44 meg:
;	partcopy fat12.bin  0   3 -f0
;	partcopy fat12.bin 26 1DA -f0 26
;
; LINUX INSTALL; 1.44 meg floppies only:
;	cat fat12.bin >/dev/fd0
;
; LINUX INSTALL; floppies other than 1.44 meg:
;	dd bs=1 if=fat12.bin skip=0  count=3   of=/dev/fd0
;	dd bs=1 if=fat12.bin skip=38 count=474 of=/dev/fd0 seek=38
;
; The volume serial number and volume label of the existing BPB
; will be disabled.
;
; Do not use RAWRITE (it will overwrite the first FAT).
; Be careful using PARTCOPY and DEBUG so you don't trash your hard disk.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; SS_NAME is the name of the second stage loader (or kernel) loaded
; by this code. SS_NAME must be in FAT format: all caps, 11 chars total,
; 8-char filename, 3-char extension, both padded on the right with spaces.
; The second stage file must be in the root directory, but it need not
; be physically contiguous ("defragmented") on the disk.
%define	SS_NAME	"STARTUP BIN"

; SS_ADDR is the physical address where the second stage will be loaded.
; It should be >= 7E00h to avoid overwriting this code, the stack or
; data used by this code, the BIOS data area (400h-4FFh), or the
; interrupt table (IVT; 0-3FFh). This value should also be a multiple
; of 512 bytes (the sector size) to prevent DMA errors when reading the
; floppy.
%define	SS_ADDR	8000h

; SS_ORG is the ORG (initial IP) value used by the second stage. After the
; second stage is loaded, this code will jump to it like this:
; JMP FAR ((SS_ADDR - SS_ORG) / 16) : SS_ORG
%define	SS_ORG	100h

; "No user-serviceable parts beyond this point" :)
; Actually, you can modify it if you want.

; BUF -> buffer for directory and FAT searches. The buffer holds two
; sectors because a FAT12 entry can straddle a sector boundary.
; The buffer is large enough for 4 Kbyte sectors; which is overkill
; for a floppy bootsector.

; .COM file; for test purposes only ("nasm -dDOS=1 ...")
; Needs lots of conventional memory. Will probably crash or freeze up
; or reboot...you've been warned.
%ifdef DOS
FS_ORG	equ	 100h	; this code at	 100h-2FFh
			; stack at	 300h-4FFh
BUF	equ	 500h	; buffer at	 500h-24FFh
FREE	equ	2500h	; free space at	2500h-
%else
BUF	equ	5A00h	; buffer at	5A00h-79FFh
			; stack at	7A00h-7BFFh
FS_ORG	equ	7C00h	; this code at	7C00h-7DFFh
; FREE	equ	7E00h	; free space at	7E00h-
%endif

; Macro to encourage NASM to use byte-offset addressing; for smaller code.
; Must load the address 'start' into register bp.
%define	VAR(x)	((x) - start) + bp

	CPU 386

	ORG FS_ORG
start:
	jmp short skip_bpb
	nop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; UNINITIALIZED VARIABLES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; These variables overlay the first part of this code, but the variables
; are not read from nor written to until we jump past this area.
; *** Store no more than 11 bytes worth of variables  ***
; *** here, else you will overwrite bytes_per_sector. ***

; (2 bytes) number of 16-byte paragraphs per cluster
para_per_cluster	EQU (start + 0)	; offset 0

; (2 bytes) number of 16-byte paragraphs per sector
para_per_sector		EQU (start + 2)	; offset 2

; (2 bytes) number of 32-byte directory entries per sector
dir_ents_per_sector	EQU (start + 4)	; offset 4

; (2 bytes) sector where the actual disk data starts
first_data_sector	EQU (start + 6)	; offset 6

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MINIMAL FAT12 BIOS PARAMETER BLOCK (BPB)
;
; 'Minimal' means just enough of a BPB so DOS/Windows:
; - recognizes this disk as a FAT12 floppy,
; - doesn't complain when you try to access the disk,
; - doesn't insist on a full format if you say "FORMAT /Q A:"
;
; Installation can use the BPB already present on your floppy disk.
; The values shown here work only with 1.44 meg disks (CHS=80:2:18)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Can't store constants in the OEM ID field because Win9x scribbles here:
; OEM ID		; offset 03h (03) - not used by this code
	db "_GEEZER_"
bytes_per_sector:	; offset 0Bh (11)
	dw 512
sectors_per_cluster:	; offset 0Dh (13)
	db 1
first_fat_sector:
num_reserved_sectors:	; offset 0Eh (14)
	dw 1
num_fats:		; offset 10h (16)
	db 2
num_root_dir_ents:	; offset 11h (17)
	dw 224
total_sectors:		; offset 13h (19) - not used by this code
	dw 18 * 2 * 80
media_id:		; offset 15h (21) - not used by this code
	db 0F0h
sectors_per_fat:	; offset 16h (22)
	dw 9
sectors_per_track:	; offset 18h (24)
	dw 18
heads:			; offset 1Ah (26)
	dw 2
partition_start:
num_hidden_sectors:	; offset 1Ch (28)
	dd 0
total_sectors_large:	; offset 20h (32) - not used by this code
	dd 0
drive_num:		; offset 24h (36)
	db 0
; reserved byte		; offset 25h (37) - not used by this code
	db 0
; extended_boot_signature: ; offset 26h (38) - not used by this code
	db 0		; =29h if VSN, volume label, and FS type present

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; CONSTANTS AND INITIALIZED VARIABLES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; which sector is in the FAT buffer
curr_sector:
	dw -1
ss_name:
	db SS_NAME

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; CODE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

skip_bpb:
	cld		; movsb auto-increments
	mov bp,start
%ifdef DOS
	mov dl,0	; A: drive
	mov sp,BUF
%else
	xor ax,ax
	mov ds,ax
	mov es,ax

; "mov ss,..." followed immediately by "mov sp,..." is atomic on all x86 CPUs
	mov ss,ax
	mov sp,bp
%endif
	mov [VAR(drive_num)],dl

; Calculate some values needed to navigate the FAT filesystem.
; Paragraphs per sector (a "paragraph" is 16 bytes):
	mov ax,[VAR(bytes_per_sector)]
	shr ax,4
	mov [VAR(para_per_sector)],ax

; paragraphs per cluster
	xor dh,dh
	mov dl,[VAR(sectors_per_cluster)]
	mul dx
	mov [VAR(para_per_cluster)],ax

; number of directory entries per sector (each entry is 32 bytes):
	mov ax,[VAR(bytes_per_sector)]
	shr ax,5
	mov [VAR(dir_ents_per_sector)],ax

; number of sectors used for root directory:
	mov ax,32
	mul word [VAR(num_root_dir_ents)]
	div word [VAR(bytes_per_sector)]
	mov di,ax
; leave this value in DI

; first sector of root directory:
	xor eax,eax
	mov al,[VAR(num_fats)]
	mul word [VAR(sectors_per_fat)]
	add ax,[VAR(num_reserved_sectors)]
; leave this value in EAX

; first sector of disk data area:
	mov dx,ax
	add dx,di
	mov [VAR(first_data_sector)],dx

; Scan root directory for the second stage file. Don't bother skipping
; deleted entries (first byte = 0E5h) or unused entries (first byte = 0).
; For the following loop:
; EAX = sector number		    BX = offset part of load address
; CX = root directory entry count   SI = pointer into loaded sector
; DI = root directory sector count  ES = segment part of load address
read_root_1:
	mov bx,BUF
	mov si,bx
	mov cx,1
	call read_sectors_chs
	jc disk_error
	mov cx,[VAR(dir_ents_per_sector)]
read_root_2:
	pusha		; save DI,SI,CX
		mov di,ss_name
		mov cx,11 ; 8.3 FAT filename
		rep cmpsb
	popa
	je found_it	; leave SI pointing to directory entry
	add si,byte 32	; bytes per FAT directory entry
	loop read_root_2
	inc eax		; next sector
	dec di
	jne read_root_1
	mov al,'F'	; file not found; display blinking 'F'

; 'Hide' the next 2-byte instruction by converting it to CMP AX,NNNN
; I learned this trick from Microsoft's Color Computer BASIC :)
	db 3Dh
disk_error:
	mov al,'R'	; disk read error; display blinking 'R'

	db 3Dh
fs_error:		; fileSystem error: corrupt FAT entries, or FAT
	mov al,'S'	; and directory entry disagree about file size

error:
	mov ah,09h	; INT 10h function: display character & attribute
	mov bx,009Fh	; BH=page=0; BL=attribute=blinking blue-on-white
	mov cx,1	; character count
	int 10h

; could omit this code to save 5 bytes:
	mov ax,0E07h	; *** BEEEP ***
	int 10h
%ifdef DOS
	mov ax,4C01h
	int 21h
%else
	mov ah,0	; await key pressed
	int 16h

	int 19h		; re-start the boot process
%endif

found_it:
%ifndef DOS
; add file size to load address, then convert to Kbytes
	mov edx,[si + 28]
	add edx,(SS_ADDR + 1023) ; +1023 to round up
	shr edx,10

; Get top of conventional RAM (Kbytes), and error if not enough memory.
; Compare all 32 bits in case of freakishly large (>= 64 meg) second stage;
; in which case (size_in_bytes >> 10) >= 0x10000L (i.e. a 32-bit value)
	xor eax,eax	; zero top 16 bits of EAX
	int 12h
	cmp eax,edx
	mov al,'M'	; not enough memory; display blinking 'M'
	jb error
%endif

; Convert file size to sector count in DI.
; Using 32-bit registers here doesn't seem to make the code any smaller.
	mov ax,[si + 28]
	mov dx,[si + 30]
	mov bx,[VAR(bytes_per_sector)]
	dec bx
	add ax,bx	; +(bytes_per_sector-1) to round up
	adc dx,byte 0
	inc bx
	div bx
	mov di,ax

; get starting cluster of file
	mov ax,[si + 26]

; SI = segment of load address
%ifdef DOS
	mov si,ds
	mov bx,FREE
%else
	mov si,(SS_ADDR / 16)
	mov bx,(SS_ADDR & 15)	; this should =0
%endif

; For the following loop:
; AX = cluster number		     BX = offset part of load address
; CX = number of sectors to load in this loop
; SI = segment part of load address  DI = current num. sectors to be loaded
	jmp short load_2
load_1:
	call fat12_lookup
	jc disk_error
load_2:
; FAT type:	FAT12	    FAT16	FAT32
; bad cluster:	  0FF7h	      0FFF7h	  0FFFFFF7h
; EOC:		>=0FF8h	    >=0FFF8h	>=0FFFFFF8h
; 0 in FAT means free cluster;
; 0 as starting cluster number means zero-length file.
	cmp ax,0FF7h
	jae fs_error
	push ax

; convert 16-bit cluster value (in AX) to 32-bit LBA sector value (in DX:AX)
		sub ax,byte 2

; cluster value of 0 or 1 here is an error:
		jc fs_error	; xxx - pop ax before taking the jump?
		xor dh,dh
		mov dl,[VAR(sectors_per_cluster)]
		mul dx
		add ax,[VAR(first_data_sector)]
		adc dx,byte 0
; DX:AX -> EAX
		push dx
		push ax
		pop eax

; CX = min(sectors_remaining_in_file, sectors_per_cluster)
; Load entire clusters until partial cluster at end of file is reached.
		xor ch,ch
		mov cl,[VAR(sectors_per_cluster)]
		cmp cx,di
		jb load_3
		mov cx,di
load_3:
; ES = segment of load address
		mov es,si
		call read_sectors_chs
	pop ax
	jc disk_error
	add si,[VAR(para_per_cluster)]
	sub di,cx
	jne load_1

; turn off floppy motor (could omit this code to save 6 bytes)
	mov dx,3F2h
	mov al,0
	out dx,al

; jump to second stage loaded at SS_ADDR and ORGed to address SS_ORG
%ifdef DOS
	xor eax,eax	; form linear address of FREE
	mov ax,ds
	shl eax,4
; +15 to round up to paragraph boundary,
; - SS_ORG to get correct CS value for far jump (retf, actually)
	add eax,(FREE + 15 - SS_ORG)
	shr eax,4
	push ax
	push word SS_ORG
	retf
%else
	jmp ((SS_ADDR - SS_ORG) / 16) : SS_ORG
%endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			read_sectors_chs
; action:		reads one or more disk sectors using INT 13h AH=02h
; in:			EAX=LBA number of sector to read, CX=sector count,
;			ES:BX -> buffer
; out (disk error):	CY=1
; out (success):	CY=0
; modifies:		(nothing)
; minimum CPU:		'386
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

read_sectors_chs:
	push es
	o32 pusha
		mov si,es

; use EDI to zero-extend 16-bit values to 32-bit
		xor edi,edi

; add number of hidden sectors (=partition start for hard disk)
		add eax,[VAR(num_hidden_sectors)]
		inc cx
		jmp short rsc_3
rsc_1:
		push cx
		push eax

; EAX=LBA sector number
; divide by number of sectors per track to get sector number
			xor edx,edx
			mov di,[VAR(sectors_per_track)]
			div edi
			mov cx,dx

; EAX/AX=quotient, CX/CL=remainder=sector (S) - 1
; divide quotient by number of heads
			xor edx,edx
			mov di,[VAR(heads)]
			div edi

; EAX/AX/AL=quotient=cylinder, DX/DL=remainder=head,
; CX/CL=sector (S) - 1. Error if cylinder >=1024:
			cmp eax,1024
			cmc
			jb rsc_2

; move variables into registers for INT 13h AH=02h
			mov es,si
			mov dh,dl	; DH	=head
			inc cl		; CL 5:0=sector
			mov ch,ah	; CH 1:0=cylinder 9:8
			shl cl,2
			shr cx,2	; CL 7:6=cylinder 9:8
			mov ch,al	; CH	=cylinder 7:0
			mov dl,[VAR(drive_num)] ; DL=drive

; INT 13h AH=02h called once for each sector because multi-sector
; reads may fail if we cross a track boundary. Seems fast enough...
			mov ax,0201h	; AH=02h, AL=num_sectors
			int 13h
			jnc rsc_2
; if error, reset drive...
			xor ax,ax
			int 13h
			jc rsc_2
; ...and try read again
			mov ax,0201h
			int 13h
rsc_2:
		pop eax
		pop cx
		jc rsc_4

; increment segment part of address and LBA sector number, and loop
		add si,[VAR(para_per_sector)]
		inc eax
rsc_3:
		loop rsc_1
rsc_4:
	o32 popa
	pop es
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			fat12_lookup
; action:		finds next cluster in chain
; in:			AX = cluster
; out (disk error):	CY=1
; out (success):	CY=0, AX = next cluster
; modifies:		AX
; minimum CPU:		'186/'286
; notes:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

fat12_lookup:
	push es
	pusha
		mov bx,BUF

; FAT12 entries are 12 bits, bytes are 8 bits. Ratio is 3 / 2,
; so multiply cluster by 3 now, and divide by 2 later.
		mov di,ax
		shl ax,1
		add ax,di

; AX b0		="HL": use high or low 12 bits of 16-bit value read from FAT
; AX b9:1	=byte offset into FAT sector (9 bits assumes 512-byte sectors)
; AX b?:10	=which sector of FAT to load
; shift "HL" into carry, then PUSHF to save it on the stack
		shr ax,1
		pushf
			xor dx,dx
			div word [VAR(bytes_per_sector)]

; remainder is byte offset into FAT sector: put it in DI
			mov di,dx
; shift "HL" into DI b0
		popf
		rcl di,1

; quotient in AX is sector-in-FAT:
; add FAT starting sector to get sector-in-disk
		add ax,[VAR(first_fat_sector)]

; check the FAT buffer to see if this sector is already loaded
; (simple disk cache; speeds things up a little --
; actually, it speeds things up a lot)
		cmp ax,[VAR(curr_sector)]
		je fl_1
		mov [VAR(curr_sector)],ax

; Read the target FAT sector. (Actually, read 2 sectors --
; FAT12 entries may straddle a sector boundary.)
		xor dx,dx
		mov cx,2

; The main load routine modifies ES. In this routine, we want
; ES to point to the original data segment:
		push ds
		pop es
		call read_sectors_chs
		jc fl_4
fl_1:
; shift "HL" from DI b0 back into carry
		shr di,1

; get 16 bits from FAT
		mov ax,[bx + di]

; look at "HL" to see if we want the high 12 bits or the low 12 bits
		jnc fl_2	; CY=0: use low 12 bits
		shr ax,4	; CY=1: use high 12 bits
fl_2:
		and ah,0Fh	; this is no smaller than "and ax,0FFFFh" :(

; clear CY bit to signal success
;		clc		; 'AND' instruction already did this
fl_4:
		mov bx,sp	; store AX to be popped by 'POPA' below
		mov [bx + 14],ax
	popa
	pop es
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; pad with NOPs to offset 510
	times (510 + $$ - $) nop

; 2-byte magic bootsector signature
	db 55h, 0AAh

